/*!
 * MAX31856_noDelay.cpp
 * Rev 2 modified for connection multiple MAX31856 chips and fast processors
 * See top of MAX31856.cpp for changing startup default settings
 *
 * The no delay mods to Adafruit's MAX31856 library are
 * by Matthew Ford
 * (c)2019 Forward Computing and Control Pty. Ltd., NSW Australia
 * All rights reserved.
 * Subject to the BSD licence below, this code may be freely used for both private and commerical use.
 * Provide this copyright is maintained.
 *
 * These mods also include code based on the multi-thermocouple code from https://forum.arduino.cc/index.php?topic=390824.0
 *
 * Note: this library does not use the MAX31856 onboard averaging.

 * @mainpage Adafruit MAX31856 thermocouple reader
 *
 * @section intro_sec Introduction
 *
 * This is the documentation for Adafruit's MAX31856 driver for the
 * Arduino platform.  It is designed specifically to work with the
 * Adafruit MAX31856 breakout: https://www.adafruit.com/products/3263
 *
 * These sensors use SPI to communicate, 4 pins are required to
 *  interface with the breakout.
 *
 * Adafruit invests time and resources providing this open source code,
 * please support Adafruit and open-source hardware by purchasing
 * products from Adafruit!
 *
 * @section dependencies Dependencies
 *
 * This library depends on <a href="https://github.com/adafruit/Adafruit_Sensor">
 * Adafruit_Sensor</a> being present on your system. Please make sure you have
 * installed the latest version before using this library.
 *
 * @section author Author
 *
 * Written by ladyada for Adafruit Industries.
 *
 * @section license License
 *
 * BSD license, all text here must be included in any redistribution.
 *
 */

#include "MAX31856_noDelay.h"
#ifdef __AVR
#include <avr/pgmspace.h>
#elif defined(ESP8266)
#include <pgmspace.h>
#endif

#include <stdlib.h>
#include <SPI.h>

static SPISettings max31856_spisettings = SPISettings(400000, MSBFIRST, SPI_MODE1); // 4Mhz clock, (max clock 5Mhz for MAX31856), clock sample high->low transition and clock is normally low
//  The  MAX31856  automatically  accommodates either clock polarity by sampling SCLK when CSbecomes  active  to  determine  the  polarity  of  the  inactive  clock.

#define MAX31856_50HZ_NOISE_FILTER_BIT 0x01
#define MAX31856_60HZ_NOISE_FILTER_BIT 0x00

#define NumRegisters 10
// edit these to change the default settings you want
static const char* RegisterNames[NumRegisters] =   {"CR0", "CR1", "MASK", "CJHF", "CHLF", "LTHFTH", "LTHFTL", "LTLFTH", "LTLFTL", "CJTO"};
static byte RegisterAddresses[NumRegisters] = {0x00,  0x01,   0x02,   0x03,   0x04,   0x05,     0x06,     0x07,     0x08,     0x09 };
static byte RegisterValues[NumRegisters] =    { //
  MAX31856_CR0_OCFAULT0 | MAX31856_60HZ_NOISE_FILTER_BIT,   // Single shot mode, do not use enum filter values here use MAX31856_50HZ_NOISE_FILTER_BIT or MAX31856_60HZ_NOISE_FILTER_BIT values
  MAX31856_TCTYPE_K, //
  0xFF, // factory default, mask all faults from aserting the fault line (not connected)
  0x7F, // factory default
  0xC0, // factory default
  0x7F,  // factory default
  0xFF,  // factory default
  0x80,  // factory default
  0x00,  // factory default
  0x00 // factory default
};

// below is a break down of the above and what the values mean
// https://datasheets.maximintegrated.com/en/ds/MAX31856.pdf   page 18
// register address 0x00 is for CR0, 0x80 = Automatic converstion, 0x10 = Open circuit fault type K, so 0x90 is both
// register address 0x01 is for CR1, 0x03 - Type K thermocouple
// register address 0x02 is for Mask, 0x02 = Mask Voltage under over fault.
// register address 0x03 is for Cold-Junction High Fault Threshold  (CJHF), 0x7F = factory default
// register address 0x04 is for Cold-Junction Low Fault Threshold  (CJLF), 0xC0 = factory default
// register address 0x05 is for Linearized Temperature High Fault Threshold MSB (LTHFTH), 0x7F = factory default
// register address 0x06 is for Linearized Temperature High Fault Threshold LSB (LTHFTL), 0xFF = factory default
// register address 0x07 is for Linearized Temperature High Fault Threshold LSB (LTLFTH), 0x80 = factory default
// register address 0x08 is for Linearized Temperature High Fault Threshold LSB (LTLFTL), 0x00 = factory default
// register address 0x09 is for Cold-Junction Temperature Offset Register (CJTO), 0x00 = factory default

/***** from MAX31856_noDelay.h

#define MAX31856_CR0_REG           0x00    ///< Config 0 register
#define MAX31856_CR0_AUTOCONVERT   0x80    ///< Config 0 Auto convert flag
#define MAX31856_CR0_1SHOT         0x40    ///< Config 0 one shot convert flag
#define MAX31856_CR0_OCFAULT1      0x20    ///< Config 0 open circuit fault 1 flag
#define MAX31856_CR0_OCFAULT0      0x10    ///< Config 0 open circuit fault 0 flag
#define MAX31856_CR0_CJ            0x08    ///< Config 0 cold junction disable flag
#define MAX31856_CR0_FAULT         0x04    ///< Config 0 fault mode flag
#define MAX31856_CR0_FAULTCLR      0x02    ///< Config 0 fault clear flag
#define MAX31856_CR0_FILTER        0x01    ///< Config 0 for 60hz filter, or 1 for 50hz filter

#define MAX31856_CR1_REG           0x01    ///< Config 1 register
#define MAX31856_MASK_REG          0x02    ///< Fault Mask register
#define MAX31856_CJHF_REG          0x03    ///< Cold junction High temp fault register
#define MAX31856_CJLF_REG          0x04    ///< Cold junction Low temp fault register
#define MAX31856_LTHFTH_REG        0x05    ///< Linearized Temperature High Fault Threshold Register, MSB
#define MAX31856_LTHFTL_REG        0x06    ///< Linearized Temperature High Fault Threshold Register, LSB
#define MAX31856_LTLFTH_REG        0x07    ///< Linearized Temperature Low Fault Threshold Register, MSB
#define MAX31856_LTLFTL_REG        0x08    ///< Linearized Temperature Low Fault Threshold Register, LSB
#define MAX31856_CJTO_REG          0x09    ///< Cold-Junction Temperature Offset Register
#define MAX31856_CJTH_REG          0x0A    ///< Cold-Junction Temperature Register, MSB
#define MAX31856_CJTL_REG          0x0B    ///< Cold-Junction Temperature Register, LSB
#define MAX31856_LTCBH_REG         0x0C    ///< Linearized TC Temperature, Byte 2
#define MAX31856_LTCBM_REG         0x0D    ///< Linearized TC Temperature, Byte 1
#define MAX31856_LTCBL_REG         0x0E    ///< Linearized TC Temperature, Byte 0
#define MAX31856_SR_REG            0x0F    ///< Fault Status Register

#define MAX31856_FAULT_CJRANGE     0x80    ///< Fault status Cold Junction Out-of-Range flag
#define MAX31856_FAULT_TCRANGE     0x40    ///< Fault status Thermocouple Out-of-Range flag
#define MAX31856_FAULT_CJHIGH      0x20    ///< Fault status Cold-Junction High Fault flag
#define MAX31856_FAULT_CJLOW       0x10    ///< Fault status Cold-Junction Low Fault flag
#define MAX31856_FAULT_TCHIGH      0x08    ///< Fault status Thermocouple Temperature High Fault flag
#define MAX31856_FAULT_TCLOW       0x04    ///< Fault status Thermocouple Temperature Low Fault flag
#define MAX31856_FAULT_OVUV        0x02    ///< Fault status Overvoltage or Undervoltage Input Fault flag
#define MAX31856_FAULT_OPEN        0x01    ///< Fault status Thermocouple Open-Circuit Fault flag


// Noise filtering options enum. Use with setNoiseFilter()
typedef enum {
MAX31856_NOISE_FILTER_60HZ,
MAX31856_NOISE_FILTER_50HZ
} max31856_noise_filter_t;

// Multiple types of thermocouples supported
typedef enum
{
  MAX31856_TCTYPE_B  = 0b0000,
  MAX31856_TCTYPE_E  = 0b0001,
  MAX31856_TCTYPE_J  = 0b0010,
  MAX31856_TCTYPE_K  = 0b0011,
  MAX31856_TCTYPE_N  = 0b0100,
  MAX31856_TCTYPE_R  = 0b0101,
  MAX31856_TCTYPE_S  = 0b0110,
  MAX31856_TCTYPE_T  = 0b0111,
  MAX31856_VMODE_G8  = 0b1000,
  MAX31856_VMODE_G32 = 0b1100,
} max31856_thermocoupletype_t;
*********/


bool MAX31856_noDelay::spiTypeSet = false;
int8_t MAX31856_noDelay::_sclk = -1;
int8_t  MAX31856_noDelay::_miso = -1;
int8_t MAX31856_noDelay::_mosi = -1;
bool MAX31856_noDelay::initSPIcalled = false;

Print* MAX31856_noDelay::debugPtr = NULL;

/**
  Set where to send debug output, if any
  Need to uncomment #define DEBUG in SpeedStepper.cpp as well as calling this method
*/
void MAX31856_noDelay::setDebugOut(Print& _debugOut) {
  debugPtr = &_debugOut;
}

void MAX31856_noDelay::initSPItype(int8_t spi_mosi, int8_t spi_miso, int8_t spi_clk) {
  if (spiTypeSet) {
    return;
  } // else
  _sclk = spi_clk;
  _miso = spi_miso;
  _mosi = spi_mosi;
  spiTypeSet = true; // set SPI type
}

/**************************************************************************/
/*!
    @brief  Instantiate MAX31856 object and use software SPI pins
    @param  spi_cs Bitbang SPI Chip Select
    @param  spi_mosi Bitbang SPI MOSI
    @param  spi_miso Bitbang SPI MISO
    @param  spi_clk Bitbang SPI Clock
*/
/**************************************************************************/
MAX31856_noDelay::MAX31856_noDelay(int8_t spi_cs, int8_t spi_mosi, int8_t spi_miso, int8_t spi_clk) {
  _cs = spi_cs;
  initSPItype(spi_mosi, spi_miso, spi_clk);
  defaultsSet = false;
}

/**************************************************************************/
/*!
    @brief  Instantiate MAX31856 object and use hardware SPI
    @param  spi_cs Any pin for SPI Chip Select
*/
/**************************************************************************/
MAX31856_noDelay::MAX31856_noDelay(int8_t spi_cs) {
  _cs = spi_cs;
  if (spiTypeSet) {
    return;
  } // else
  initSPItype(-1,-1,-1);
  defaultsSet = false;
}

/**************************************************************************/
/*!
    @brief  Initialize MAX31856 attach/set pins or SPI device, default to K thermocouple
    @returns Always returns true at this time (no known way of detecting chip ID)
*/
/**************************************************************************/
boolean MAX31856_noDelay::begin(void) {
  pinMode(_cs, OUTPUT);
  digitalWrite(_cs, HIGH);

  if (!initSPIcalled) {
    initSPIcalled = true;
    if (_sclk != -1) {
      //define pin modes
      pinMode(_sclk, OUTPUT);
      digitalWrite(_sclk,LOW);
      pinMode(_mosi, OUTPUT);
      digitalWrite(_mosi,LOW);
      pinMode(_miso, INPUT);
    } else {
      //start and configure hardware SPI
      SPI.begin();
    }
  }
  return true;
}

// reset all settings to defaults
// this call blocks for 200ms to set noise filter
void MAX31856_noDelay::resetDefaults() {
  defaultsSet = false;
  setDefaults();
}

// this call blocks for 200ms to set noise filter
void MAX31856_noDelay::setDefaults() {
  if (defaultsSet) {
    return;
  }
  defaultsSet = true;
  if (debugPtr) {
    debugPtr->print(F("Initializing MAX31856 on CS pin "));
    debugPtr->println(_cs);
  }
  for (int i = 0; i < NumRegisters; i++) {
    writeRegister8(RegisterAddresses[i], RegisterValues[i]);
  }
  // fix up filter setting, need to turn off autoconvert to set this.
  uint8_t t = readRegister8(MAX31856_CR0_REG);
  uint8_t filter50Hz = t& MAX31856_CR0_FILTER;
  if (filter50Hz) {
  	setNoiseFilter(MAX31856_NOISE_FILTER_50HZ);
  } else {
  	setNoiseFilter(MAX31856_NOISE_FILTER_60HZ);
  }

  verifyDefaults();
}

void MAX31856_noDelay::verifyDefaults() {
  int ErrorCount = 0;
  for (int i = 0; i < NumRegisters; i++) {
    byte RegVal = readRegister8(RegisterAddresses[i]);
    if (RegVal != RegisterValues[i]) {
      if (debugPtr) {
        debugPtr->print(RegisterNames[i]);
        debugPtr->print(F("\t has 0x"));
        debugPtr->print(RegVal, HEX);
        debugPtr->print(F(" and should have 0x"));
        debugPtr->println(RegisterValues[i], HEX);
      }
      ErrorCount++;
    }
  }
  if (ErrorCount == 0) {
    if (debugPtr) {
      debugPtr->println(F("Initialization of defaults successful"));
    }
  }
}

/**************************************************************************/
/*!
    @brief  Set which kind of Thermocouple (K, J, T, etc) to detect & decode
    @param type The enumeration type of the thermocouple
*/
/**************************************************************************/
void MAX31856_noDelay::setThermocoupleType(max31856_thermocoupletype_t type) {
  setDefaults();
  uint8_t t = readRegister8(MAX31856_CR1_REG);
  t &= 0xF0; // mask off bottom 4 bits
  t |= (uint8_t)type & 0x0F;
  writeRegister8(MAX31856_CR1_REG, t);
}

/**************************************************************************/
/*!
    @brief  Get which kind of Thermocouple (K, J, T, etc) we are using
    @returns The enumeration type of the thermocouple
*/
/**************************************************************************/
max31856_thermocoupletype_t MAX31856_noDelay::getThermocoupleType(void) {
  setDefaults();
  uint8_t t = readRegister8(MAX31856_CR1_REG);
  t &= 0x0F;

  return (max31856_thermocoupletype_t)(t);
}

/**************************************************************************/
/*!
    @brief  Read the fault register (8 bits)
    @returns 8 bits of fault register data
*/
/**************************************************************************/
uint8_t MAX31856_noDelay::readFault(void) {
  setDefaults();
  return readRegister8(MAX31856_SR_REG);
}

/**************************************************************************/
/*!
    @brief  Sets the threshhold for internal chip temperature range
    for fault detection. NOT the thermocouple temperature range!
    @param  low Low (min) temperature, signed 8 bit so -128 to 127 degrees C
    @param  high High (max) temperature, signed 8 bit so -128 to 127 degrees C
*/
/**************************************************************************/
void MAX31856_noDelay::setColdJunctionFaultThreshholds(int8_t low, int8_t high) {
  setDefaults();
  writeRegister8(MAX31856_CJLF_REG, low);
  writeRegister8(MAX31856_CJHF_REG, high);
}

/**************************************************************************/
/*!
    @brief  Sets the mains noise filter. Can be set to 50 or 60hz.
    NOTE: this call blocks for 200ms to allow last autoconvert conversion to complete (if any)
    Defaults to 60hz. You need to call this if you live in a 50hz country.
    @param  noiseFilter One of MAX31856_NOISE_FILTER_50HZ or MAX31856_NOISE_FILTER_60HZ
*/
/**************************************************************************/
void MAX31856_noDelay::setNoiseFilter(max31856_noise_filter_t noiseFilter) {
  setDefaults();
  uint8_t init_CR0 = readRegister8(MAX31856_CR0_REG);
  // turn off autoconvert
  oneShotTemperature();
  delay(200); // wait for last conversion to complete
  // set filter
  uint8_t t = readRegister8(MAX31856_CR0_REG);
  if(noiseFilter == MAX31856_NOISE_FILTER_50HZ) {
    t|=0x01;
  } else {
    t&=0xfe;
  }
  writeRegister8(MAX31856_CR0_REG, t);
  // restore previous setting for autoconvert but with this new filter settting
  t = init_CR0;
  if(noiseFilter == MAX31856_NOISE_FILTER_50HZ) {
    t|=0x01;
  } else {
    t&=0xfe;
  }
  writeRegister8(MAX31856_CR0_REG, t);
}

/**************************************************************************/
/*!
    @brief  Sets the threshhold for thermocouple temperature range
    for fault detection. NOT the internal chip temperature range!
    @param  flow Low (min) temperature, floating point
    @param  fhigh High (max) temperature, floating point
*/
/**************************************************************************/
void MAX31856_noDelay::setTempFaultThreshholds(float flow, float fhigh) {
  setDefaults();
  int16_t low, high;

  flow *= 16;
  low = flow;

  fhigh *= 16;
  high = fhigh;

  writeRegister8(MAX31856_LTHFTH_REG, high >> 8);
  writeRegister8(MAX31856_LTHFTL_REG, high);

  writeRegister8(MAX31856_LTLFTH_REG, low >> 8);
  writeRegister8(MAX31856_LTLFTL_REG, low);
}

/**************************************************************************
  Set to autoconvert 
  Need to call this after using oneshot OR if default settings are for one shot
  Value must be read later, not returned here!
*/
void MAX31856_noDelay::setAutoConvert(void) {
  setDefaults();

  uint8_t t = readRegister8(MAX31856_CR0_REG);

  t |= MAX31856_CR0_AUTOCONVERT; // set off autoconvert!

  writeRegister8(MAX31856_CR0_REG, t); // conversion starts at end of this write
  // check for result in  ~ every 100ms if no averaging set
}

/**************************************************************************/
/*!
    @brief  Begin a one-shot (read temperature only upon request) measurement.
    Value must be read later, not returned here!
*/
/**************************************************************************/
void MAX31856_noDelay::oneShotTemperature(void) {
  setDefaults();

  uint8_t t = readRegister8(MAX31856_CR0_REG);

  t &= ~MAX31856_CR0_AUTOCONVERT; // turn off autoconvert!
  t |= MAX31856_CR0_1SHOT; //set 1 for one shot

  writeRegister8(MAX31856_CR0_REG, t); // conversion starts at end of this write
  // check for result in  ~ 143ms in 60Hz filter mode or ~169ms in 50Hz filter mode
  // oneshot bit self clears so need to call this again to start next conversion
}

/**************************************************************************/
/*!
    @brief  Start a one-shot measurement and return internal chip temperature
    @returns Floating point temperature of chip in Celsius
*/
/**************************************************************************/
float MAX31856_noDelay::readCJTemperature(void) {
  setDefaults();

  int16_t temp16 = readRegister16(MAX31856_CJTH_REG);
  float tempfloat = temp16;
  tempfloat /= 256.0;

  return tempfloat;
}

/**************************************************************************/
/*!
    @brief  Start a one-shot measurement and return thermocouple tip temperature
    @returns Floating point temperature at end of thermocouple in Celsius
*/
/**************************************************************************/
float MAX31856_noDelay::readThermocoupleTemperature(void) {
  setDefaults();

  int32_t temp24 = readRegister24(MAX31856_LTCBH_REG);
  if (temp24 & 0x800000) {
    temp24 |= 0xFF000000;  // fix sign
  }

  temp24 >>= 5;  // bottom 5 bits are unused

  float tempfloat = temp24;
  tempfloat *= 0.0078125;

  return tempfloat;
}


uint8_t MAX31856_noDelay::readRegister8(uint8_t addr) {
  uint8_t ret = 0;
  readRegisterN(addr, &ret, 1);

  return ret;
}

uint16_t MAX31856_noDelay::readRegister16(uint8_t addr) {
  uint8_t buffer[2] = {0, 0};
  readRegisterN(addr, buffer, 2);

  uint16_t ret = buffer[0];
  ret <<= 8;
  ret |=  buffer[1];

  return ret;
}

uint32_t MAX31856_noDelay::readRegister24(uint8_t addr) {
  uint8_t buffer[3] = {0, 0, 0};
  readRegisterN(addr, buffer, 3);

  uint32_t ret = buffer[0];
  ret <<= 8;
  ret |=  buffer[1];
  ret <<= 8;
  ret |=  buffer[2];

  return ret;
}


void MAX31856_noDelay::readRegisterN(uint8_t addr, uint8_t buffer[], uint8_t n) {

  addr &= 0x7F; // make sure top bit is not set

  if (_sclk == -1) {
    SPI.beginTransaction(max31856_spisettings);
  } else {
    digitalWrite(_sclk, HIGH);
  }

  digitalWrite(_cs, LOW);
  delayMicroseconds(1);

  spixfer(addr);
  delayMicroseconds(1);

  while (n--) {
    buffer[0] = spixfer(0xFF);
    buffer++;
  }

  if (_sclk == -1) {
    SPI.endTransaction();
  }
  delayMicroseconds(1); // clock to CS hold delay 100nS min
  digitalWrite(_cs, HIGH);
}


void MAX31856_noDelay::writeRegister8(uint8_t addr, uint8_t data) {
  addr |= 0x80; // make sure top bit is set

  if (_sclk == -1) {
    SPI.beginTransaction(max31856_spisettings);
  } else {
    digitalWrite(_sclk, HIGH);
  }

  digitalWrite(_cs, LOW);
  delayMicroseconds(1);

  spixfer(addr);
  delayMicroseconds(1);
  spixfer(data);

  if (_sclk == -1) {
    SPI.endTransaction();
  }
  delayMicroseconds(1);

  digitalWrite(_cs, HIGH);
}



uint8_t MAX31856_noDelay::spixfer(uint8_t x) {
  if (_sclk == -1) {
    return SPI.transfer(x);
  }

  // software spi
  uint8_t reply = 0;
  for (int i=7; i>=0; i--) {
    reply <<= 1;
    digitalWrite(_sclk, LOW);
    digitalWrite(_mosi, x & (1<<i));
    digitalWrite(_sclk, HIGH);
    if (digitalRead(_miso)) {
      reply |= 1;
    }
  }
  return reply;
}
